@@ -132,6 +132,7 @@ tpm2tss_tpm2data_write(const TPM2_DATA *tpm2Data, const char *filename)
     TSS2_RC r;
     BIO *bio = NULL;
     TSSPRIVKEY *tpk = NULL;
+    BIGNUM *bn_parent = NULL;
 
     uint8_t privbuf[sizeof(tpm2Data->priv)];
     uint8_t pubbuf[sizeof(tpm2Data->pub)];
@@ -170,12 +171,17 @@ tpm2tss_tpm2data_write(const TPM2_DATA *tpm2Data, const char *filename)
         goto error;
     }
 
-    tpk->emptyAuth = ! !tpm2Data->emptyAuth;
-    if (tpm2Data->parent != 0) {
-        ASN1_INTEGER_set(tpk->parent, tpm2Data->parent);
-    } else {
-        ASN1_INTEGER_set(tpk->parent, TPM2_RH_OWNER);
+    tpk->emptyAuth = tpm2Data->emptyAuth ? 0xFF : 0;
+    bn_parent = BN_new();
+    if (!bn_parent) {
+        goto error;
     }
+    if (tpm2Data->parent != 0) {
+        BN_set_word(bn_parent, tpm2Data->parent);
+    } else {
+        BN_set_word(bn_parent, TPM2_RH_OWNER);
+    }
+    BN_to_ASN1_INTEGER(bn_parent, tpk->parent);
     ASN1_STRING_set(tpk->privkey, &privbuf[0], privbuf_len);
     ASN1_STRING_set(tpk->pubkey, &pubbuf[0], pubbuf_len);
 
@@ -312,7 +318,7 @@ session_error:
 
     esys_ctx_free(&esys_ctx);
     tpm2Data->pub = *outPublic;
-    free(outPublic);
+    Esys_Free(outPublic);
 
     *tpm2Datap = tpm2Data;
     return 1;
@@ -341,6 +347,7 @@ tpm2tss_tpm2data_read(const char *filename, TPM2_DATA **tpm2Datap)
     TSSPRIVKEY *tpk = NULL;
     TPM2_DATA *tpm2Data = NULL;
     char type_oid[64];
+    BIGNUM *bn_parent;
 
     if ((bio = BIO_new_file(filename, "r")) == NULL) {
         ERR(tpm2tss_tpm2data_read, TPM2TSS_R_FILE_READ);
@@ -364,9 +371,17 @@ tpm2tss_tpm2data_read(const char *filename, TPM2_DATA **tpm2Datap)
 
     tpm2Data->privatetype = KEY_TYPE_BLOB;
 
-    tpm2Data->emptyAuth = tpk->emptyAuth;
+    tpm2Data->emptyAuth = !!tpk->emptyAuth;
 
-    tpm2Data->parent = ASN1_INTEGER_get(tpk->parent);
+    bn_parent = ASN1_INTEGER_to_BN(tpk->parent, NULL);
+    if (!bn_parent) {
+        goto error;
+    }
+    if (BN_is_negative(bn_parent)) {
+        tpm2Data->parent = ASN1_INTEGER_get(tpk->parent);
+    } else {
+        tpm2Data->parent = BN_get_word(bn_parent);
+    }
     if (tpm2Data->parent == 0)
         tpm2Data->parent = TPM2_RH_OWNER;
 
@@ -482,6 +497,24 @@ init_tpm_parent(ESYS_CONTEXT **esys_ctx,
         }
     }
 
+    /*
+     * TPM2_ALG_ECC is *mandatory* for TPM2.0; the above should never
+     * fail. However, *if* such a broken TPM is used then ephemeral
+     * primaries according to the TSS2 PEM file standard can *never*
+     * have worked on that hardware, so it isn't *breaking* anything
+     * for us to unilaterally use an ephemeral RSA parent in this case
+     * instead.
+     *
+     * However, it may not be interoperable to do so, and it isn't a
+     * good idea anyway since RSA keys are *slow* to generate, so
+     * users with a broken TPM like this really *should* have followed
+     * the recommendation to create the RSA primary and store it in
+     * the NVRAM at 0x81000001. And then the TSS2 PEM keys should use
+     * *that* as the parent, not the ephemeral version. In fact, there
+     * is a strong case to be made for defaulting to 0x81000001 if it
+     * exists, *before* (or never) falling back to generating an RSA
+     * key here.
+     */
     if (primaryTemplate == NULL) {
         for (index = 0; index < capabilityData->data.algorithms.count; index++) {
             if (capabilityData->data.algorithms.algProperties[index].alg == TPM2_ALG_RSA) {
@@ -491,8 +524,7 @@ init_tpm_parent(ESYS_CONTEXT **esys_ctx,
         }
     }
 
-    if (capabilityData != NULL)
-        free (capabilityData);
+    Esys_Free (capabilityData);
 
     if (primaryTemplate == NULL) {
         ERR(init_tpm_parent, TPM2TSS_R_UNKNOWN_ALG);
@@ -592,3 +624,75 @@ init_tpm_key (ESYS_CONTEXT **esys_ctx, ESYS_TR *keyHandle, TPM2_DATA *tpm2Data)
     esys_ctx_free(esys_ctx);
     return r;
 }
+
+/** Deserialize a tpm key from disk
+ *
+ * Read a tpm key as marshaled TPM2B_PUBLIC and (encrypted) TPM2B_PRIVATE from
+ * disk and convert them into a TPM2_DATA representation
+ * @param filenamepub The filename to read the public portion from.
+ * @param filenametpm The filename to read the private portion from.
+ * @param parent Handle of the parent key.
+ * @param emptyAuth Whether the object does not require authentication.
+ * @param tpm2Datap The data after read.
+ * @retval 1 on success
+ * @retval 0 on failure
+ */
+int
+tpm2tss_tpm2data_importtpm(const char *filenamepub, const char *filenametpm,
+                           TPM2_HANDLE parent, int emptyAuth,
+                           TPM2_DATA **tpm2Datap)
+{
+    TSS2_RC r;
+    BIO *bio;
+    TPM2_DATA *tpm2data;
+    int filepub_size, filepriv_size;
+
+    uint8_t filepub[sizeof(TPM2B_PUBLIC)];
+    uint8_t filepriv[sizeof(TPM2B_PRIVATE)];
+
+    if ((bio = BIO_new_file(filenamepub, "r")) == NULL) {
+        ERR(tpm2tss_tpm2data_read, TPM2TSS_R_FILE_READ);
+        return 0;
+    }
+    filepub_size = BIO_read(bio, &filepub[0], sizeof(filepub));
+    BIO_free(bio);
+    if (filepub_size < 0) {
+        ERR(tpm2tss_tpm2data_read, TPM2TSS_R_FILE_READ);
+        return 0;
+    }
+
+    if ((bio = BIO_new_file(filenametpm, "r")) == NULL) {
+        ERR(tpm2tss_tpm2data_read, TPM2TSS_R_FILE_READ);
+        return 0;
+    }
+    filepriv_size = BIO_read(bio, &filepriv[0], sizeof(filepriv));
+    BIO_free(bio);
+    if (filepriv_size < 0) {
+        ERR(tpm2tss_tpm2data_read, TPM2TSS_R_FILE_READ);
+        return 0;
+    }
+
+    tpm2data = OPENSSL_malloc(sizeof(TPM2_DATA));
+    if (!tpm2data)
+        return 0;
+
+    memset(tpm2data, 0, sizeof(*tpm2data));
+    tpm2data->privatetype = KEY_TYPE_BLOB;
+    tpm2data->parent = parent;
+    tpm2data->emptyAuth = emptyAuth;
+
+    r = Tss2_MU_TPM2B_PUBLIC_Unmarshal(&filepub[0], filepub_size, NULL,
+                                       &tpm2data->pub);
+    ERRchktss(tpm2tss_tpm2data_read, r, goto error);
+
+    r = Tss2_MU_TPM2B_PRIVATE_Unmarshal(&filepriv[0], filepriv_size, NULL,
+                                        &tpm2data->priv);
+    ERRchktss(tpm2tss_tpm2data_read, r, goto error);
+
+    *tpm2Datap = tpm2data;
+    return 1;
+
+  error:
+    OPENSSL_free(tpm2data);
+    return 0;
+}
