@@ -48,6 +48,13 @@ RSA_METHOD rsa_methods;
 RSA_METHOD *rsa_methods = NULL;
 #endif /* OPENSSL_VERSION_NUMBER < 0x10100000 */
 
+#ifdef HAVE_OPENSSL_DIGEST_SIGN
+static int (*rsa_pkey_orig_copy)(EVP_PKEY_CTX *dst, EVP_PKEY_CTX *src);
+static void (*rsa_pkey_orig_cleanup)(EVP_PKEY_CTX *ctx);
+#endif /* HAVE_OPENSSL_DIGEST_SIGN */
+
+static int (*rsa_orig_finish)(RSA *rsa);
+
 static TPM2B_DATA allOutsideInfo = {
     .size = 0,
 };
@@ -169,6 +176,10 @@ rsa_priv_enc(int flen, const unsigned char *from, unsigned char *to, RSA *rsa,
     DBGBUF(&sig->buffer[0], sig->size);
 
     ret = sig->size;
+    if (ret > RSA_size(rsa) || ret <= 0) {
+        ERR(rsa_priv_enc, TPM2TSS_R_DIGEST_TOO_LARGE);
+        goto error;
+    }
     memcpy(to, &sig->buffer[0], ret);
 
     goto out;
@@ -177,7 +188,7 @@ rsa_priv_enc(int flen, const unsigned char *from, unsigned char *to, RSA *rsa,
     r = -1;
 
  out:
-    free(sig);
+    Esys_Free(sig);
     if (keyHandle != ESYS_TR_NONE) {
         if (tpm2Data->privatetype == KEY_TYPE_HANDLE) {
             Esys_TR_Close(esys_ctx, &keyHandle);
@@ -226,7 +237,7 @@ rsa_priv_dec(int flen, const unsigned char *from, unsigned char *to, RSA * rsa,
     TPMT_RSA_DECRYPT inScheme;
 
     TPM2B_PUBLIC_KEY_RSA cipher = { .size = flen };
-    if (flen > (int)sizeof(cipher.buffer)) {
+    if (flen > (int)sizeof(cipher.buffer) || flen < 0) {
         ERR(rsa_priv_dec, TPM2TSS_R_DIGEST_TOO_LARGE);
         goto error;
     }
@@ -240,6 +251,9 @@ rsa_priv_dec(int flen, const unsigned char *from, unsigned char *to, RSA * rsa,
         inScheme.scheme = TPM2_ALG_OAEP;
         inScheme.details.oaep.hashAlg = TPM2_ALG_SHA1;
         break;
+    case RSA_NO_PADDING:
+        inScheme.scheme = TPM2_ALG_NULL;
+        break;
     default:
         ERR(rsa_priv_dec, TPM2TSS_R_PADDING_UNKNOWN);
         goto error;
@@ -257,6 +271,10 @@ rsa_priv_dec(int flen, const unsigned char *from, unsigned char *to, RSA * rsa,
     DBGBUF(&message->buffer[0], message->size);
 
     flen = message->size;
+    if (flen > RSA_size(rsa) || flen <= 0) {
+        ERR(rsa_priv_dec, TPM2TSS_R_DIGEST_TOO_LARGE);
+        goto error;
+    }
     memcpy(to, &message->buffer[0], flen);
 
     goto out;
@@ -265,7 +283,7 @@ rsa_priv_dec(int flen, const unsigned char *from, unsigned char *to, RSA * rsa,
     r = -1;
 
  out:
-    free(message);
+    Esys_Free(message);
     if (keyHandle != ESYS_TR_NONE) {
         if (tpm2Data->privatetype == KEY_TYPE_HANDLE) {
             Esys_TR_Close(esys_ctx, &keyHandle);
@@ -278,6 +296,26 @@ rsa_priv_dec(int flen, const unsigned char *from, unsigned char *to, RSA * rsa,
     return (r == TSS2_RC_SUCCESS) ? flen : 0;
 }
 
+/** Clean up the RSA key
+ *
+ * @param rsa The rsa key object.
+ * @retval 1 on success, or 0 on failure
+ */
+static int
+rsa_finish(RSA *rsa)
+{
+    TPM2_DATA *tpm2Data = RSA_get_app_data(rsa);
+
+    if (tpm2Data != NULL) {
+        OPENSSL_free(tpm2Data);
+        RSA_set_app_data(rsa, NULL);
+    }
+    if (rsa_orig_finish) {
+       rsa_orig_finish(rsa);
+    }
+    return 1;
+}
+
 /** Helper to populate the RSA key object.
  *
  * In order to use an RSA key object in a typical manner, all fields of the
@@ -415,8 +453,9 @@ populate_rsa(RSA *rsa)
  *
  * This function creates a key object given a TPM2_DATA object. The resulting
  * key object can then be used for signing and decrypting with the tpm2tss
- * engine.
- * @param tpm2Data The key data to use.
+ * engine. Ownership of the TPM2_DATA object is taken on success.
+ * @param tpm2Data The key data to use. Must have been allocated using
+ * OPENSSL_malloc.
  * @retval key The key object
  * @retval NULL on failure.
  */
@@ -456,8 +495,10 @@ tpm2tss_rsa_makekey(TPM2_DATA *tpm2Data)
         goto error;
     }
 
-    if (!populate_rsa(rsa))
+    if (!populate_rsa(rsa)) {
+        RSA_set_app_data(rsa, NULL);
         goto error;
+    }
 
     DBG("Created RSA key object.\n");
 
@@ -558,12 +599,14 @@ tpm2tss_rsa_genkey(RSA *rsa, int bits, BIGNUM *e, char *password,
     goto end;
  error:
     r = -1;
+    if (rsa)
+        RSA_set_app_data(rsa, NULL);
     if (tpm2Data)
         OPENSSL_free(tpm2Data);
 
  end:
-    free(keyPrivate);
-    free(keyPublic);
+    Esys_Free(keyPrivate);
+    Esys_Free(keyPublic);
 
     if (parent != ESYS_TR_NONE && !parentHandle)
         Esys_FlushContext(esys_ctx, parent);
@@ -592,6 +635,113 @@ RSA_METHOD rsa_methods = {
 };
 #endif                          /* OPENSSL_VERSION_NUMBER < 0x10100000 */
 
+#ifdef HAVE_OPENSSL_DIGEST_SIGN
+static int
+rsa_pkey_copy(EVP_PKEY_CTX *dst, EVP_PKEY_CTX *src)
+{
+    if (rsa_pkey_orig_copy && !rsa_pkey_orig_copy(dst, src))
+        return 0;
+
+    return digest_sign_copy(dst, src);
+}
+
+static void
+rsa_pkey_cleanup(EVP_PKEY_CTX *ctx)
+{
+    digest_sign_cleanup(ctx);
+
+    if (rsa_pkey_orig_cleanup)
+        rsa_pkey_orig_cleanup(ctx);
+}
+
+/* called for digest & sign init, after message digest algorithm set */
+static int
+rsa_digest_custom(EVP_PKEY_CTX *ctx, EVP_MD_CTX *mctx)
+{
+    EVP_PKEY *pkey = EVP_PKEY_CTX_get0_pkey(ctx);
+    RSA *rsa = EVP_PKEY_get0_RSA(pkey);
+    TPM2_DATA *tpm2data = RSA_get_app_data(rsa);
+
+    DBG("rsa_digest_custom %p %p\n", ctx, mctx);
+
+    return digest_sign_init(ctx, mctx, tpm2data, RSA_size(rsa));
+}
+
+static int
+rsa_signctx(EVP_PKEY_CTX *ctx, unsigned char *sig, size_t *siglen,
+            EVP_MD_CTX *mctx)
+{
+    TPM2_SIG_DATA *sig_data = EVP_PKEY_CTX_get_app_data(ctx);
+    TSS2_RC r = TSS2_RC_SUCCESS;
+    TPMT_TK_HASHCHECK *validation_ptr = NULL;
+    TPM2B_DIGEST *digest_ptr = NULL;
+    TPMT_SIGNATURE *tpm_sig = NULL;
+    int pad_mode;
+
+    DBG("rsa_signctx %p %p sig_data %p\n", ctx, mctx, sig_data);
+
+    if (!sig) {
+        /* caller just wants to know the size */
+        *siglen = sig_data->sig_size;
+        return 1;
+    }
+
+    if (!sig_data) {
+        /* handle non-TPM key */
+        unsigned char md[EVP_MAX_MD_SIZE];
+        unsigned int md_len = 0;
+
+        if (!EVP_DigestFinal_ex(mctx, md, &md_len))
+            return 0;
+        if (EVP_PKEY_sign(ctx, sig, siglen, md, md_len) <= 0)
+            return 0;
+        return 1;
+    }
+
+    if (EVP_PKEY_CTX_get_rsa_padding(ctx, &pad_mode) <= 0)
+        return 0;
+
+    TPMT_SIG_SCHEME in_scheme = {
+        .scheme = TPM2_ALG_NULL,
+        .details.rsassa.hashAlg = sig_data->hash_alg,
+    };
+    switch (pad_mode) {
+    case RSA_PKCS1_PADDING:
+        in_scheme.scheme = TPM2_ALG_RSASSA;
+        break;
+    case RSA_PKCS1_PSS_PADDING:
+        in_scheme.scheme = TPM2_ALG_RSAPSS;
+        break;
+    default:
+        ERR(rsa_signctx, TPM2TSS_R_PADDING_UNKNOWN);
+        return 0;
+    }
+
+    if (!digest_finish(sig_data, &digest_ptr, &validation_ptr))
+        return 0;
+
+    r = Esys_Sign(sig_data->key->esys_ctx, sig_data->key->key_handle,
+                  ESYS_TR_PASSWORD, ESYS_TR_NONE, ESYS_TR_NONE,
+                  digest_ptr, &in_scheme, validation_ptr, &tpm_sig);
+    ERRchktss(rsa_signctx, r, goto error);
+
+    memcpy(sig, tpm_sig->signature.rsassa.sig.buffer, sig_data->sig_size);
+    *siglen = sig_data->sig_size;
+
+    r = 1;
+    goto out;
+
+ error:
+    r = 0;
+ out:
+    Esys_Free(tpm_sig);
+    Esys_Free(digest_ptr);
+    Esys_Free(validation_ptr);
+
+    return r;
+}
+#endif /* HAVE_OPENSSL_DIGEST_SIGN */
+
 /** Initialize the tpm2tss engine's rsa submodule
  *
  * Initialize the tpm2tss engine's submodule by setting function pointer.
@@ -612,7 +762,8 @@ init_rsa(ENGINE *e)
     rsa_methods.rsa_mod_exp = default_rsa->rsa_mod_exp;
     rsa_methods.bn_mod_exp = default_rsa->bn_mod_exp;
 
-    return ENGINE_set_RSA(e, &rsa_methods);
+    if (!ENGINE_set_RSA(e, &rsa_methods))
+        return 0;
 #else /* OPENSSL_VERSION_NUMBER < 0x10100000 */
     default_rsa = RSA_PKCS1_OpenSSL();
     if (default_rsa == NULL)
@@ -622,7 +773,41 @@ init_rsa(ENGINE *e)
     RSA_meth_set1_name(rsa_methods, "TPM2TSS RSA methods");
     RSA_meth_set_priv_enc(rsa_methods, rsa_priv_enc);
     RSA_meth_set_priv_dec(rsa_methods, rsa_priv_dec);
+    rsa_orig_finish = RSA_meth_get_finish(rsa_methods);
+    RSA_meth_set_finish(rsa_methods, rsa_finish);
 
-    return ENGINE_set_RSA(e, rsa_methods);
+    if (!ENGINE_set_RSA(e, rsa_methods))
+        return 0;
 #endif /* OPENSSL_VERSION_NUMBER < 0x10100000 */
+
+#if HAVE_OPENSSL_DIGEST_SIGN
+    /* digest and sign support */
+
+    EVP_PKEY_METHOD *pkey_rsa_methods;
+
+    pkey_rsa_methods = EVP_PKEY_meth_new(EVP_PKEY_RSA,
+                                         EVP_PKEY_FLAG_AUTOARGLEN);
+    if (pkey_rsa_methods == NULL)
+        return 0;
+
+    const EVP_PKEY_METHOD *pkey_orig_rsa_methods =
+        EVP_PKEY_meth_find(EVP_PKEY_RSA);
+    if (pkey_orig_rsa_methods == NULL)
+        return 0;
+    EVP_PKEY_meth_copy(pkey_rsa_methods, pkey_orig_rsa_methods);
+    /*
+     * save originals since we only override some of the pkey
+     * functionality, rather than reimplementing all of it
+     */
+    EVP_PKEY_meth_get_copy(pkey_rsa_methods, &rsa_pkey_orig_copy);
+    EVP_PKEY_meth_get_cleanup(pkey_rsa_methods, &rsa_pkey_orig_cleanup);
+
+    EVP_PKEY_meth_set_copy(pkey_rsa_methods, rsa_pkey_copy);
+    EVP_PKEY_meth_set_cleanup(pkey_rsa_methods, rsa_pkey_cleanup);
+    EVP_PKEY_meth_set_signctx(pkey_rsa_methods, NULL, rsa_signctx);
+    EVP_PKEY_meth_set_digest_custom(pkey_rsa_methods, rsa_digest_custom);
+    EVP_PKEY_meth_add0(pkey_rsa_methods);
+#endif /* HAVE_OPENSSL_DIGEST_SIGN */
+
+    return 1;
 }
